/******************************************************************************************************************************************
 *
 * Filename    : mem1.c
 * Project     : MemAlloc
 * Description : This is an example implementation of the C memory functions malloc(), free(), calloc() and realloc().
 *               It should serve as an tutorial for learning - it's not an efficient, optimized implementation.
 *
 *               See the corresponding article on my homepage for explanation of the algorithms.
 *
 *               This file contains the implementation variant 1 of the C memory functions.
 *
 * Sunshine, October 2k19
 * www.sunshine2k.de | www.bastian-molkenthin.de
 *
 * This file is licensed under MIT licence. For further details, see http://www.sunshine2k.de/license.html.
 *
 *****************************************************************************************************************************************/

/******************************************************************************************************************************************
 * INCLUDES
 *****************************************************************************************************************************************/
#include "mem1.h"
#if defined ( MEM1_TEST_MODE )
# include <stdio.h>
# include "testutil.h"
#endif

/******************************************************************************************************************************************
 * DEFINES
 *****************************************************************************************************************************************/
#define HEAP_TOTAL_SIZE       4*1024
#define HEAP_NODE_SIZE        sizeof(MyHeapNodeType)

/******************************************************************************************************************************************
 * TYPES
 *****************************************************************************************************************************************/
typedef struct MyHeapNodeTypeTag
{
  uint32_t size;                              /* size of heap, without size of control data */
  uint8_t  used;                              /* allocated by application or not*/
  struct MyHeapNodeTypeTag* nextNode;         /* pointer to next heap entry */
  struct MyHeapNodeTypeTag* prevNode;         /* pointer to next heap entry */
} MyHeapNodeType;

/******************************************************************************************************************************************
 * VARIABLES
 *****************************************************************************************************************************************/
static unsigned char MyHeapArea[HEAP_TOTAL_SIZE];
static MyHeapNodeType* heapStart = (MyHeapNodeType *)MyHeapArea;

/******************************************************************************************************************************************
 * LOCAL FUNCTION DECLARATIONS
 *****************************************************************************************************************************************/
#if defined ( MEM1_TEST_MODE )
static void mem1_printHeapheader();
#endif


/******************************************************************************************************************************************
 * FUNCTIONS
 *****************************************************************************************************************************************/

/**************************************************************************************************
 * mem1_init
 **************************************************************************************************/
void mem1_init()
{
  heapStart = (MyHeapNodeType *)MyHeapArea;
  heapStart->size = HEAP_TOTAL_SIZE - HEAP_NODE_SIZE;
  heapStart->nextNode = NULL_PTR;
  heapStart->prevNode = NULL_PTR;
}

/**************************************************************************************************
 * mem1_malloc
 **************************************************************************************************/
void * mem1_malloc(size_t size)
{
  MyHeapNodeType* currentHeapBlock;
  MyHeapNodeType* bestHeapBlock;
  uint32_t bestHeapBlockSize;

  /* init current block to start of heap */
  currentHeapBlock = heapStart;
  /* init best heap block */
  bestHeapBlock = (MyHeapNodeType*)NULL_PTR;
  bestHeapBlockSize = HEAP_TOTAL_SIZE + 1; /* init with invalid size */

  while (currentHeapBlock)
  {
    /* check if current block matches and fittest best (so far) */
    if ( (!currentHeapBlock->used) && 
         (currentHeapBlock->size >= (size + HEAP_NODE_SIZE)) && 
         (currentHeapBlock->size <= bestHeapBlockSize) )
    {
      bestHeapBlock = currentHeapBlock;
      bestHeapBlockSize = currentHeapBlock->size;
    }

    currentHeapBlock = currentHeapBlock->nextNode;
  }

  if (bestHeapBlock != NULL_PTR)
  {
    MyHeapNodeType* heapNodeAllocate;

    /* found a matching block, split it up and return the top of the memory area to the user */
    /* the best matching block is decreased by the needed memory area*/
    bestHeapBlock->size = bestHeapBlock->size - size - HEAP_NODE_SIZE;
    /* new heap node is after the current heap + the size of its control data + allocated memory size */
    heapNodeAllocate = (MyHeapNodeType*)(((unsigned char*)bestHeapBlock) + HEAP_NODE_SIZE + bestHeapBlock->size);
    heapNodeAllocate->size = size;
    heapNodeAllocate->used = 1;
    heapNodeAllocate->nextNode = bestHeapBlock->nextNode;
    heapNodeAllocate->prevNode = bestHeapBlock;
    if (bestHeapBlock->nextNode != NULL_PTR)
    {
      /* next block exists */
      bestHeapBlock->nextNode->prevNode = heapNodeAllocate;
    }
    bestHeapBlock->nextNode = heapNodeAllocate;
    /* return pointer to memory of new heap node after control data */
    return (void*)((unsigned char*)heapNodeAllocate + HEAP_NODE_SIZE);
  }

  return NULL_PTR;
}

/**************************************************************************************************
 * mem1_free
 **************************************************************************************************/
void mem1_free(void* p)
{
  if (p == NULL_PTR)
  {
    return;
  }

  /* get actual heap node */
  MyHeapNodeType* currentBlock = (MyHeapNodeType*)((unsigned char*)p - HEAP_NODE_SIZE);

  if (currentBlock == NULL_PTR)
  {
    return;
  }

  currentBlock->used = 0;

  /* check if we can merge with next block */
  if (currentBlock->nextNode != NULL_PTR)
  {
    if (!currentBlock->nextNode->used)
    {
      /* add size of next block and its control data to current block */
      currentBlock->size += currentBlock->nextNode->size;
      currentBlock->size += HEAP_NODE_SIZE;

      /* remove next block */
      /* link current block to next-next block */
      currentBlock->nextNode = currentBlock->nextNode->nextNode;
      /* link next-next block to current block if next-next block exists */
      if (currentBlock->nextNode != NULL_PTR) /* currentBlock->nextNode points to next-next block already! */
      {
        currentBlock->nextNode->prevNode = currentBlock;
      }
    }
  }

  /* check if we can merge with previous block */
  if (currentBlock->prevNode != NULL_PTR)
  {
    if (!currentBlock->prevNode->used)
    {
      /* add size of freed memory block and its control data to previous block */
      currentBlock->prevNode->size += currentBlock->size;
      currentBlock->prevNode->size += HEAP_NODE_SIZE;

      /* remove freed block from list */
      /* link previous block to next block */
      currentBlock->prevNode->nextNode = currentBlock->nextNode;
      /* link next block to previous block if next block exists */
      if (currentBlock->nextNode != NULL_PTR)
      {
        currentBlock->nextNode->prevNode = currentBlock->prevNode;
      }
    }
  }
}

/**************************************************************************************************
 * mem1_calloc
 **************************************************************************************************/
void* mem1_calloc(size_t num, size_t size)
{
  void* p;
  size_t sizeInBytes;

  /* calculate actual memory size in bytes */
  sizeInBytes = num * size;
  /* allocate memory block */
  p = mem1_malloc(sizeInBytes);
  /* initialize memory block to zero */
  if (p != NULL_PTR)
  {
    mem1_memset(p, 0, sizeInBytes);
  }
  return p;
}

/**************************************************************************************************
 * mem1_realloc
 **************************************************************************************************/
void* mem1_realloc(void* ptr, size_t size)
{
  void* p;

  /* allocate new block of requested size */
  p = mem1_malloc(size);
  if (p == NULL_PTR)
  {
    return p; /* return NULL_PTR in case of failure, not original ptr! */
  }

  /* copy buffer of original block */
  if (ptr != NULL_PTR)
  {
    /* get original size of memory block */
    size_t sizeToCopy;
    size_t originalSize = ((MyHeapNodeType* )((unsigned char*)ptr - HEAP_NODE_SIZE))->size;

    /* copy data of original memory block to new one - only valid data is copied */
    if (originalSize < size)
    {
      sizeToCopy = originalSize;
    }
    else
    {
      sizeToCopy = size;
    }
    mem1_memcpy(p, ptr, sizeToCopy);

    /* free original block */
    mem1_free(ptr);
  }

  return p;
}

/**************************************************************************************************
 * mem1_memset
 **************************************************************************************************/
void* mem1_memset(void * ptr, int value, size_t num)
{
  uint8_t* p = (uint8_t*) ptr;
  while (num > 0)
  {
    *p = value;
    p++;
    num--;
  }
  return ptr;
}

/**************************************************************************************************
 * mem1_memcpy
 **************************************************************************************************/
void* mem1_memcpy(void * dest, const void * source, size_t num)
{
  size_t i;
  uint8_t* pDest = (uint8_t*)dest;
  uint8_t* pSource = (uint8_t*)source;

  for (i = 0; i < num; i++)
  {
    pDest[i] = pSource[i];
  }

  return dest;
}

#if defined ( MEM1_TEST_MODE )
/**************************************************************************************************
 * mem1_printHeapheader
 **************************************************************************************************/
static void mem1_printHeapheader()
{
  printf("====================================================\n");
  printf("Start of Heap: 0x%08X\n", (unsigned int)heapStart);
  printf("Size of Heap : 0x%08X (%d)\n", HEAP_TOTAL_SIZE - HEAP_NODE_SIZE, HEAP_TOTAL_SIZE - HEAP_NODE_SIZE);
  printf("====================================================\n");
}

/**************************************************************************************************
 * mem1_printHeap
 **************************************************************************************************/
void mem1_printHeap()
{
  MyHeapNodeType* currentHeapBlock;
  printf("\n--------------------------------\n");
  printf("Start\t\tUsed\tSize\t\t\tPrev\t\tNext\n");

  currentHeapBlock = heapStart;
  while (currentHeapBlock)
  {
    printf("0x%08X\t0x%02X\t0x%08X (%d)\t0x%08X\t0x%08X \n", 
      (unsigned int)currentHeapBlock, currentHeapBlock->used, currentHeapBlock->size, currentHeapBlock->size,
      (unsigned int)currentHeapBlock->prevNode, (unsigned int)currentHeapBlock->nextNode);

    currentHeapBlock = currentHeapBlock->nextNode;
  }

}

/**************************************************************************************************
 * mem1_printHeapIfFailed
 **************************************************************************************************/
void mem1_printHeapIfFailed(void)
{
  if (test_get_testcase_failed() != 0)
  {
    mem1_printHeap();
  }
}

/**************************************************************************************************
 * mem1_getSizeOfHeapCtrlNode
 **************************************************************************************************/
uint32_t  mem1_getSizeOfHeapCtrlNode()
{
  return HEAP_NODE_SIZE;
}

/**************************************************************************************************
 * mem1_getSizeOfTotalHeapMem
 **************************************************************************************************/
uint32_t  mem1_getSizeOfTotalHeapMem()
{
  return HEAP_TOTAL_SIZE;
}


/**************************************************************************************************
 * mem1_getNumberOfNodes
 **************************************************************************************************/
uint8_t   mem1_getNumberOfNodes(void)
{
  uint8_t numOfNodes = 1;
  MyHeapNodeType* node = heapStart;

  while (node->nextNode != NULL_PTR)
  {
    numOfNodes++;
    node = node->nextNode;
  }

  return numOfNodes;
}

/**************************************************************************************************
 * mem1_getHeapUseOfNode
 **************************************************************************************************/
uint8_t   mem1_getHeapUseOfNode(uint8_t numOfNode)
{
  uint8_t currentNodeNum = 1;
  MyHeapNodeType* node = heapStart;

  while (currentNodeNum < numOfNode && node->nextNode != NULL_PTR)
  {
    currentNodeNum++;
    node = node->nextNode;
  }

  if (currentNodeNum != numOfNode)
  {
    return 0xFF;
  }

  return node->used;
}

/**************************************************************************************************
 * mem1_getHeapSizeOfNode
 **************************************************************************************************/
uint32_t   mem1_getHeapSizeOfNode(uint8_t numOfNode)
{
  uint8_t currentNodeNum = 1;
  MyHeapNodeType* node = heapStart;

  while (currentNodeNum < numOfNode && node->nextNode != NULL_PTR)
  {
    currentNodeNum++;
    node = node->nextNode;
  }

  if (currentNodeNum != numOfNode)
  {
    return 0;
  }

  return node->size;
}

/**************************************************************************************************
 * mem1_isHeapMemorySizeValid
 **************************************************************************************************/
uint8_t mem1_isHeapMemorySizeValid(void)
{
  uint8_t numOfNodes = 0;
  MyHeapNodeType* node = heapStart;
  uint32_t size = 0;

  do
  {
    numOfNodes++;
    size += node->size;
    node = node->nextNode;
  }
  while (node != NULL_PTR);

  size = size + (numOfNodes * HEAP_NODE_SIZE);

  if (size == HEAP_TOTAL_SIZE)
  {
    return 1;
  }
  else
  {
    return 0;
  }
}
#endif
