/******************************************************************************************************************************************
 *
 * Filename    : mem1test.c
 * Project     : MemAlloc
 * Description : This is an example implementation of the C memory functions malloc(), free(), calloc() and realloc().
 *               It should serve as an tutorial for learning - it's not an efficient, optimized implementation.
 *
 *               See the corresponding article on my homepage for explanation of the algorithms.
 *
 *               This file contains the test cases for implementation variant 1 C memory functions.
 *
 * Sunshine, October 2k19
 * www.sunshine2k.de | www.bastian-molkenthin.de
 *
 * This file is licensed under MIT licence. For further details, see http://www.sunshine2k.de/license.html.
 *
 *****************************************************************************************************************************************/

/******************************************************************************************************************************************
 * INCLUDES
 *****************************************************************************************************************************************/
#include "mem1test.h"
#include "testutil.h"

/******************************************************************************************************************************************
 * DEFINES
 *****************************************************************************************************************************************/

/******************************************************************************************************************************************
 * TYPES
 *****************************************************************************************************************************************/

/******************************************************************************************************************************************
 * LOCAL FUNCTION DECLARATIONS
 *****************************************************************************************************************************************/
#if defined ( MEM1_TEST_MODE )
static void mem1_assertHeapEmpty();
#endif

/******************************************************************************************************************************************
 * LOCAL FUNCTIONS
 *****************************************************************************************************************************************/
#if defined ( MEM1_TEST_MODE )
static void mem1_assertHeapEmpty()
{
  test_assert_equal(mem1_getNumberOfNodes(), 1);
  test_assert_equal(mem1_getHeapUseOfNode(1), 0);
  test_assert_equal(mem1_getHeapSizeOfNode(1), mem1_getSizeOfTotalHeapMem() - mem1_getSizeOfHeapCtrlNode());
  test_assert_equal(mem1_isHeapMemorySizeValid(), 1);
}
#endif

/******************************************************************************************************************************************
 * FUNCTIONS
 *****************************************************************************************************************************************/

#if defined ( MEM1_TEST_MODE )
/*******************************************************
 * mem1test_FullHeapAlloc
 *
 * Steps:
 * - Allocate full heap and assert success
 * - Allocate another block and assert that it fails
 * - Free allocated block and assert that full heap is free
 *******************************************************/
void mem1test_FullHeapAlloc(void)
{
  void* p, *pFail;

  test_testcase_start("mem1test_FullHeapAlloc");
  mem1_init();
  test_assert_equal(mem1_isHeapMemorySizeValid(), 1);
  mem1_assertHeapEmpty();

  /* allocate complete heap */
  p = mem1_malloc(mem1_getSizeOfTotalHeapMem() - 2 * mem1_getSizeOfHeapCtrlNode());
  test_assert_not_null_ptr(p);

  /* assert heap */
  test_assert_equal(mem1_getNumberOfNodes(), 2);
  test_assert_equal(mem1_getHeapUseOfNode(1), 0);
  test_assert_equal(mem1_getHeapUseOfNode(2), 1);
  test_assert_equal(mem1_getHeapSizeOfNode(1), 0);
  test_assert_equal(mem1_getHeapSizeOfNode(2), mem1_getSizeOfTotalHeapMem() - 2 * mem1_getSizeOfHeapCtrlNode());
  test_assert_equal(mem1_isHeapMemorySizeValid(), 1);
  mem1_printHeapIfFailed();

  /* get another heap memory block and check that it's fail */
  pFail = mem1_malloc(mem1_getSizeOfTotalHeapMem() - mem1_getSizeOfHeapCtrlNode());
  test_assert_null_ptr(pFail);

  /* free complete allocated memory block */
  mem1_free(p);
  mem1_assertHeapEmpty();
  test_assert_equal(mem1_isHeapMemorySizeValid(), 1);

  test_testcase_end();
}

/*******************************************************
 * mem1test_FullHeapAllocInSecondRequest
 *
 * Steps:
 * - Allocate one block of 100
 * - Allocate remaining heap
 * - free block with remaining heap
 * - free block with 100
 *******************************************************/
void mem1test_FullHeapAllocInSecondRequest(void)
{
  void* p1, *p2;

  test_testcase_start("mem1test_FullHeapAllocInSecondRequest");
  mem1_init();

  /* allocate block 1 */
  p1 = mem1_malloc(100);

  /* allocate remaining heap */
  p2 = mem1_malloc(mem1_getSizeOfTotalHeapMem() - 3*mem1_getSizeOfHeapCtrlNode() - 100);
  test_assert_not_null_ptr(p2);
  test_assert_equal(mem1_getNumberOfNodes(), 3);
  test_assert_equal(mem1_getHeapUseOfNode(1), 0);
  test_assert_equal(mem1_getHeapUseOfNode(2), 1);
  test_assert_equal(mem1_getHeapUseOfNode(3), 1);
  test_assert_equal(mem1_getHeapSizeOfNode(1), 0);
  test_assert_equal(mem1_getHeapSizeOfNode(2), mem1_getSizeOfTotalHeapMem() - 3 * mem1_getSizeOfHeapCtrlNode() - 100);
  test_assert_equal(mem1_getHeapSizeOfNode(3), 100);
  test_assert_equal(mem1_isHeapMemorySizeValid(), 1);
  mem1_printHeapIfFailed();

  /* free remaing block */
  mem1_free(p2);
  test_assert_equal(mem1_getNumberOfNodes(), 2);
  test_assert_equal(mem1_getHeapUseOfNode(1), 0);
  test_assert_equal(mem1_getHeapUseOfNode(2), 1);
  test_assert_equal(mem1_getHeapSizeOfNode(1), mem1_getSizeOfTotalHeapMem() - 2 * mem1_getSizeOfHeapCtrlNode() - 100);
  test_assert_equal(mem1_getHeapSizeOfNode(2), 100);
  mem1_printHeapIfFailed();

  /* free block 1 */
  mem1_free(p1);
  mem1_assertHeapEmpty();

  test_testcase_end();
}

/*******************************************************
 * mem1test_FullHeapAllocInSecondRequest_FreeReversedOrder
 *
 * Steps:
 * - Allocate one block of 100
 * - Allocate remaining heap
 * - free block with 100
 * - free block with remaining heap
 *******************************************************/
void mem1test_FullHeapAllocInSecondRequest_FreeReversedOrder(void)
{
  void* p1, *p2;

  test_testcase_start("mem1test_FullHeapAllocInSecondRequest_FreeReversedOrder");
  mem1_init();

  /* allocate block 1 */
  p1 = mem1_malloc(100);

  /* allocate remaining heap */
  p2 = mem1_malloc(mem1_getSizeOfTotalHeapMem() - 3*mem1_getSizeOfHeapCtrlNode() - 100);
  test_assert_not_null_ptr(p2);
  test_assert_equal(mem1_getNumberOfNodes(), 3);
  test_assert_equal(mem1_getHeapUseOfNode(1), 0);
  test_assert_equal(mem1_getHeapUseOfNode(2), 1);
  test_assert_equal(mem1_getHeapUseOfNode(3), 1);
  test_assert_equal(mem1_getHeapSizeOfNode(1), 0);
  test_assert_equal(mem1_getHeapSizeOfNode(2), mem1_getSizeOfTotalHeapMem() - 3 * mem1_getSizeOfHeapCtrlNode() - 100);
  test_assert_equal(mem1_getHeapSizeOfNode(3), 100);
  test_assert_equal(mem1_isHeapMemorySizeValid(), 1);
  mem1_printHeapIfFailed();

  /* free block 1 */
  mem1_free(p1);
  test_assert_equal(mem1_getNumberOfNodes(), 3);
  test_assert_equal(mem1_getHeapUseOfNode(1), 0);
  test_assert_equal(mem1_getHeapUseOfNode(2), 1);
  test_assert_equal(mem1_getHeapUseOfNode(3), 0);
  test_assert_equal(mem1_getHeapSizeOfNode(1), 0);
  test_assert_equal(mem1_getHeapSizeOfNode(2), mem1_getSizeOfTotalHeapMem() - 3 * mem1_getSizeOfHeapCtrlNode() - 100);
  test_assert_equal(mem1_getHeapSizeOfNode(3), 100);
  test_assert_equal(mem1_isHeapMemorySizeValid(), 1);
  mem1_printHeapIfFailed();

  /* free block 2 */
  mem1_free(p2);
  mem1_assertHeapEmpty();
  mem1_printHeapIfFailed();

  test_testcase_end();
}

/*******************************************************
 * mem1test_AllocRequestSizeTooLarge
 *
 * Steps:
 * - Allocate block with too large size
 *******************************************************/
void mem1test_AllocRequestSizeTooLarge(void)
{
  void* p1, *p2;

  test_testcase_start("mem1test_AllocRequestSizeTooLarge");
  mem1_init();

  /* try to allocate block with too large requested size */
  p1 = mem1_malloc(mem1_getSizeOfTotalHeapMem() - 2 * mem1_getSizeOfHeapCtrlNode() + 1);
  test_assert_null_ptr(p1);
  mem1_assertHeapEmpty();

  /* allocate valid block */
  p1 = mem1_malloc(99);
  test_assert_equal(mem1_getNumberOfNodes(), 2);
  test_assert_equal(mem1_getHeapUseOfNode(1), 0);
  test_assert_equal(mem1_getHeapUseOfNode(2), 1);
  test_assert_equal(mem1_getHeapSizeOfNode(1), mem1_getSizeOfTotalHeapMem() - 2 * mem1_getSizeOfHeapCtrlNode() - 99);
  test_assert_equal(mem1_getHeapSizeOfNode(2), 99);
  mem1_printHeapIfFailed();

  /* try to allocate block with too large requested size */
  p2 = mem1_malloc(mem1_getSizeOfTotalHeapMem() - 3 * mem1_getSizeOfHeapCtrlNode() -99 + 1);
  test_assert_null_ptr(p2);
  test_assert_equal(mem1_getNumberOfNodes(), 2);
  test_assert_equal(mem1_getHeapUseOfNode(1), 0);
  test_assert_equal(mem1_getHeapUseOfNode(2), 1);
  test_assert_equal(mem1_getHeapSizeOfNode(1), mem1_getSizeOfTotalHeapMem() - 2 * mem1_getSizeOfHeapCtrlNode() - 99);
  test_assert_equal(mem1_getHeapSizeOfNode(2), 99);
  mem1_printHeapIfFailed();

  /* free valid block */
  mem1_free(p1);
  mem1_assertHeapEmpty();
  mem1_printHeapIfFailed();

  test_testcase_end();
}

/*******************************************************
 * mem1test_FreeBasic
 *
 * Steps:
 * - Allocate one block of 1000
 * - free block with 1000
 *******************************************************/
void mem1test_FreeBasic(void)
{
  void* p1;

  test_testcase_start("mem1test_FreeBasic");
  mem1_init();

  /* allocate block 1 */
  p1 = mem1_malloc(1000);
  test_assert_not_null_ptr(p1);
  test_assert_equal(mem1_getNumberOfNodes(), 2);
  test_assert_equal(mem1_getHeapUseOfNode(1), 0);
  test_assert_equal(mem1_getHeapUseOfNode(2), 1);
  test_assert_equal(mem1_getHeapSizeOfNode(2), 1000);
  test_assert_equal(mem1_isHeapMemorySizeValid(), 1);
  mem1_printHeapIfFailed();

  /* free block 1 */
  mem1_free(p1);
  mem1_assertHeapEmpty();

  mem1_printHeapIfFailed();
  test_testcase_end();
}

/*******************************************************
 * mem1test_FreeLatestToFirst
 *
 * Steps:
 * - Allocate one block of 100
 * - Allocate one block of 200
 * - Allocate one block of 300
 * - free block with 300
 * - free block with 200
 * - free block with 100
 *******************************************************/
void mem1test_FreeLatestToFirst()
{
  void *p1, *p2, *p3;

  test_testcase_start("mem1test_FreeLatestToFirst");
  mem1_init();

  /* allocate block 1 */
  p1 = mem1_malloc(100);
  test_assert_not_null_ptr(p1);
  test_assert_equal(mem1_getNumberOfNodes(), 2);
  test_assert_equal(mem1_getHeapUseOfNode(1), 0);
  test_assert_equal(mem1_getHeapUseOfNode(2), 1);
  test_assert_equal(mem1_getHeapSizeOfNode(2), 100);
  test_assert_equal(mem1_isHeapMemorySizeValid(), 1);
  mem1_printHeapIfFailed();

  /* allocate block 2 */
  p2 = mem1_malloc(200);
  test_assert_not_null_ptr(p2);
  test_assert_equal(mem1_getNumberOfNodes(), 3);
  test_assert_equal(mem1_getHeapUseOfNode(1), 0);
  test_assert_equal(mem1_getHeapUseOfNode(2), 1);
  test_assert_equal(mem1_getHeapUseOfNode(3), 1);
  test_assert_equal(mem1_getHeapSizeOfNode(2), 200);
  test_assert_equal(mem1_getHeapSizeOfNode(3), 100);
  test_assert_equal(mem1_isHeapMemorySizeValid(), 1);
  mem1_printHeapIfFailed();

  /* allocate block 3 */
  p3 = mem1_malloc(300);
  test_assert_not_null_ptr(p3);
  test_assert_equal(mem1_getNumberOfNodes(), 4);
  test_assert_equal(mem1_getHeapUseOfNode(1), 0);
  test_assert_equal(mem1_getHeapUseOfNode(2), 1);
  test_assert_equal(mem1_getHeapUseOfNode(3), 1);
  test_assert_equal(mem1_getHeapUseOfNode(4), 1);
  test_assert_equal(mem1_getHeapSizeOfNode(2), 300);
  test_assert_equal(mem1_getHeapSizeOfNode(3), 200);
  test_assert_equal(mem1_getHeapSizeOfNode(4), 100);
  test_assert_equal(mem1_isHeapMemorySizeValid(), 1);
  mem1_printHeapIfFailed();
  //mem1_printHeap();

  /* free block 3 */
  mem1_free(p3);
  test_assert_equal(mem1_getNumberOfNodes(), 3);
  test_assert_equal(mem1_getHeapUseOfNode(1), 0);
  test_assert_equal(mem1_getHeapUseOfNode(2), 1);
  test_assert_equal(mem1_getHeapUseOfNode(3), 1);
  test_assert_equal(mem1_getHeapSizeOfNode(2), 200);
  test_assert_equal(mem1_getHeapSizeOfNode(3), 100);
  test_assert_equal(mem1_isHeapMemorySizeValid(), 1);
  mem1_printHeapIfFailed();
  //mem1_printHeap();

  /* free block 2 */
  mem1_free(p2);
  test_assert_equal(mem1_getNumberOfNodes(), 2);
  test_assert_equal(mem1_getHeapUseOfNode(1), 0);
  test_assert_equal(mem1_getHeapUseOfNode(2), 1);
  test_assert_equal(mem1_getHeapSizeOfNode(2), 100);
  test_assert_equal(mem1_isHeapMemorySizeValid(), 1);
  mem1_printHeapIfFailed();

  /* free block 1*/
  mem1_free(p1);
  mem1_assertHeapEmpty();
  mem1_printHeapIfFailed();

  test_testcase_end();
}

/*******************************************************
 * mem1test_FreeFirstToLatest
 *
 * Steps:
 * - Allocate one block of 100
 * - Allocate one block of 200
 * - Allocate one block of 300
 * - free block with 100
 * - free block with 200
 * - free block with 300
 *******************************************************/
void mem1test_FreeFirstToLatest()
{
  void *p1, *p2, *p3;

  test_testcase_start("mem1test_FreeFirstToLatest");
  mem1_init();

  /* allocate block 1 */
  p1 = mem1_malloc(100);
  test_assert_not_null_ptr(p1);
  test_assert_equal(mem1_getNumberOfNodes(), 2);
  test_assert_equal(mem1_getHeapUseOfNode(1), 0);
  test_assert_equal(mem1_getHeapUseOfNode(2), 1);
  test_assert_equal(mem1_getHeapSizeOfNode(2), 100);
  test_assert_equal(mem1_isHeapMemorySizeValid(), 1);
  mem1_printHeapIfFailed();

  /* allocate block 2 */
  p2 = mem1_malloc(200);
  test_assert_not_null_ptr(p2);
  test_assert_equal(mem1_getNumberOfNodes(), 3);
  test_assert_equal(mem1_getHeapUseOfNode(1), 0);
  test_assert_equal(mem1_getHeapUseOfNode(2), 1);
  test_assert_equal(mem1_getHeapUseOfNode(3), 1);
  test_assert_equal(mem1_getHeapSizeOfNode(2), 200);
  test_assert_equal(mem1_getHeapSizeOfNode(3), 100);
  test_assert_equal(mem1_isHeapMemorySizeValid(), 1);
  mem1_printHeapIfFailed();

  /* allocate block 3 */
  p3 = mem1_malloc(300);
  test_assert_not_null_ptr(p3);
  test_assert_equal(mem1_getNumberOfNodes(), 4);
  test_assert_equal(mem1_getHeapUseOfNode(1), 0);
  test_assert_equal(mem1_getHeapUseOfNode(2), 1);
  test_assert_equal(mem1_getHeapUseOfNode(3), 1);
  test_assert_equal(mem1_getHeapUseOfNode(4), 1);
  test_assert_equal(mem1_getHeapSizeOfNode(2), 300);
  test_assert_equal(mem1_getHeapSizeOfNode(3), 200);
  test_assert_equal(mem1_getHeapSizeOfNode(4), 100);
  test_assert_equal(mem1_isHeapMemorySizeValid(), 1);
  mem1_printHeapIfFailed();
  //mem1_printHeap();

  /* free block 1 */
  mem1_free(p1);
  test_assert_equal(mem1_getNumberOfNodes(), 4);
  test_assert_equal(mem1_getHeapUseOfNode(1), 0);
  test_assert_equal(mem1_getHeapUseOfNode(2), 1);
  test_assert_equal(mem1_getHeapUseOfNode(3), 1);
  test_assert_equal(mem1_getHeapUseOfNode(4), 0);
  test_assert_equal(mem1_getHeapSizeOfNode(2), 300);
  test_assert_equal(mem1_getHeapSizeOfNode(3), 200);
  test_assert_equal(mem1_getHeapSizeOfNode(4), 100);
  test_assert_equal(mem1_isHeapMemorySizeValid(), 1);
  mem1_printHeapIfFailed();
  //mem1_printHeap();

  /* free block 2 */
  mem1_free(p2);
  //mem1_printHeap();
  test_assert_equal(mem1_getNumberOfNodes(), 3);
  test_assert_equal(mem1_getHeapUseOfNode(1), 0);
  test_assert_equal(mem1_getHeapUseOfNode(2), 1);
  test_assert_equal(mem1_getHeapUseOfNode(3), 0);
  test_assert_equal(mem1_getHeapSizeOfNode(2), 300);
  test_assert_equal(mem1_getHeapSizeOfNode(3), 316);
  test_assert_equal(mem1_isHeapMemorySizeValid(), 1);
  mem1_printHeapIfFailed();
 // mem1_printHeap();

  /* free block 3 */
  mem1_free(p3);
  mem1_assertHeapEmpty();
  mem1_printHeapIfFailed();

  test_testcase_end();
}

/*******************************************************
 * mem1test_FreeMiddlePrevious
 *
 * Steps:
 * - Allocate one block of 100
 * - Allocate one block of 200
 * - Allocate one block of 300
 * - free block with 200
 * - free block with 300
 * - free block with 100
 *******************************************************/
void mem1test_FreeMiddlePrevious()
{
  void *p1, *p2, *p3;

  test_testcase_start("mem1test_FreeMiddlePrevious");
  mem1_init();

  /* allocate block 1 */
  p1 = mem1_malloc(100);
  /* allocate block 2 */
  p2 = mem1_malloc(200);
  /* allocate block 3 */
  p3 = mem1_malloc(300);

  /* free block 2 */
  mem1_free(p2);
  test_assert_equal(mem1_getNumberOfNodes(), 4);
  test_assert_equal(mem1_getHeapUseOfNode(1), 0);
  test_assert_equal(mem1_getHeapUseOfNode(2), 1);
  test_assert_equal(mem1_getHeapUseOfNode(3), 0);
  test_assert_equal(mem1_getHeapUseOfNode(4), 1);
  test_assert_equal(mem1_getHeapSizeOfNode(2), 300);
  test_assert_equal(mem1_getHeapSizeOfNode(3), 200);
  test_assert_equal(mem1_getHeapSizeOfNode(4), 100);
  test_assert_equal(mem1_isHeapMemorySizeValid(), 1);
  mem1_printHeapIfFailed();

  /* free block 3 */
  mem1_free(p3);
  test_assert_equal(mem1_getNumberOfNodes(), 2);
  test_assert_equal(mem1_getHeapUseOfNode(1), 0);
  test_assert_equal(mem1_getHeapUseOfNode(2), 1);
  test_assert_equal(mem1_getHeapSizeOfNode(1), mem1_getSizeOfTotalHeapMem() - 2*mem1_getSizeOfHeapCtrlNode() - 100);
  test_assert_equal(mem1_getHeapSizeOfNode(2), 100);
  test_assert_equal(mem1_isHeapMemorySizeValid(), 1);
  mem1_printHeapIfFailed();

  /* free block 1 */
  mem1_free(p1);
  mem1_assertHeapEmpty();
  mem1_printHeapIfFailed();

  test_testcase_end();
}

/*******************************************************
 * mem1test_FreeMiddleNext
 *
 * Steps:
 * - Allocate one block of 100
 * - Allocate one block of 200
 * - Allocate one block of 300
 * - free block with 200
 * - free block with 100
 * - free block with 300
 *******************************************************/
void mem1test_FreeMiddleNext()
{
  void *p1, *p2, *p3;

  test_testcase_start("mem1test_FreeMiddleNext");
  mem1_init();

  /* allocate block 1 */
  p1 = mem1_malloc(100);
  /* allocate block 2 */
  p2 = mem1_malloc(200);
  /* allocate block 3 */
  p3 = mem1_malloc(300);
  //mem1_printHeap();

  /* free block 2 */
  mem1_free(p2);
  test_assert_equal(mem1_getNumberOfNodes(), 4);
  test_assert_equal(mem1_getHeapUseOfNode(1), 0);
  test_assert_equal(mem1_getHeapUseOfNode(2), 1);
  test_assert_equal(mem1_getHeapUseOfNode(3), 0);
  test_assert_equal(mem1_getHeapUseOfNode(4), 1);
  test_assert_equal(mem1_getHeapSizeOfNode(2), 300);
  test_assert_equal(mem1_getHeapSizeOfNode(3), 200);
  test_assert_equal(mem1_getHeapSizeOfNode(4), 100);
  test_assert_equal(mem1_isHeapMemorySizeValid(), 1);
  mem1_printHeapIfFailed();

  /* free block 1 */
  mem1_free(p1);
  test_assert_equal(mem1_getNumberOfNodes(), 3);
  test_assert_equal(mem1_getHeapUseOfNode(1), 0);
  test_assert_equal(mem1_getHeapUseOfNode(2), 1);
  test_assert_equal(mem1_getHeapUseOfNode(3), 0);
  test_assert_equal(mem1_getHeapSizeOfNode(1), mem1_getSizeOfTotalHeapMem() - 3*mem1_getSizeOfHeapCtrlNode() - 300 - 316);
  test_assert_equal(mem1_getHeapSizeOfNode(2), 300);
  test_assert_equal(mem1_getHeapSizeOfNode(3), 316);
  test_assert_equal(mem1_isHeapMemorySizeValid(), 1);
  mem1_printHeapIfFailed();

  /* free block 3 */
  mem1_free(p3);
  mem1_assertHeapEmpty();
  mem1_printHeapIfFailed();

  test_testcase_end();
}

/*******************************************************
 * mem1test_FreeMiddleTwoPrevious
 *
 * Steps:
 * - Allocate one block of 100
 * - Allocate one block of 200
 * - Allocate one block of 300
 * - Allocate one block of 400
 * - free block with 200
 * - free block with 300
 * - free block with 400
 * - free block with 100
 *******************************************************/
void mem1test_FreeMiddleTwoPrevious()
{
  void *p1, *p2, *p3, *p4;

  test_testcase_start("mem1test_FreeMiddleTwoPrevious");
  mem1_init();

  /* allocate block 1 */
  p1 = mem1_malloc(100);
  /* allocate block 2 */
  p2 = mem1_malloc(200);
  /* allocate block 3 */
  p3 = mem1_malloc(300);
  /* allocate block 4 */
  p4 = mem1_malloc(400);

  test_assert_not_null_ptr(p4);
  test_assert_equal(mem1_getNumberOfNodes(), 5);
  test_assert_equal(mem1_getHeapUseOfNode(1), 0);
  test_assert_equal(mem1_getHeapUseOfNode(2), 1);
  test_assert_equal(mem1_getHeapUseOfNode(3), 1);
  test_assert_equal(mem1_getHeapUseOfNode(4), 1);
  test_assert_equal(mem1_getHeapUseOfNode(5), 1);
  test_assert_equal(mem1_getHeapSizeOfNode(2), 400);
  test_assert_equal(mem1_getHeapSizeOfNode(3), 300);
  test_assert_equal(mem1_getHeapSizeOfNode(4), 200);
  test_assert_equal(mem1_getHeapSizeOfNode(5), 100);
  test_assert_equal(mem1_isHeapMemorySizeValid(), 1);
  mem1_printHeapIfFailed();

  /* free block 2 */
  mem1_free(p2);
  test_assert_equal(mem1_getNumberOfNodes(), 5);
  test_assert_equal(mem1_getHeapUseOfNode(1), 0);
  test_assert_equal(mem1_getHeapUseOfNode(2), 1);
  test_assert_equal(mem1_getHeapUseOfNode(3), 1);
  test_assert_equal(mem1_getHeapUseOfNode(4), 0);
  test_assert_equal(mem1_getHeapUseOfNode(5), 1);
  test_assert_equal(mem1_getHeapSizeOfNode(2), 400);
  test_assert_equal(mem1_getHeapSizeOfNode(3), 300);
  test_assert_equal(mem1_getHeapSizeOfNode(4), 200);
  test_assert_equal(mem1_getHeapSizeOfNode(5), 100);
  test_assert_equal(mem1_isHeapMemorySizeValid(), 1);
  mem1_printHeapIfFailed();

  /* free block 3 */
  mem1_free(p3);
  test_assert_equal(mem1_getNumberOfNodes(), 4);
  test_assert_equal(mem1_getHeapUseOfNode(1), 0);
  test_assert_equal(mem1_getHeapUseOfNode(2), 1);
  test_assert_equal(mem1_getHeapUseOfNode(3), 0);
  test_assert_equal(mem1_getHeapUseOfNode(4), 1);
  test_assert_equal(mem1_getHeapSizeOfNode(2), 400);
  test_assert_equal(mem1_getHeapSizeOfNode(3), 300 + 200 + 16);
  test_assert_equal(mem1_getHeapSizeOfNode(4), 100);
  test_assert_equal(mem1_isHeapMemorySizeValid(), 1);
  mem1_printHeapIfFailed();

  /* free block 4 */
  mem1_free(p4);
  test_assert_equal(mem1_getNumberOfNodes(), 2);
  test_assert_equal(mem1_getHeapUseOfNode(1), 0);
  test_assert_equal(mem1_getHeapUseOfNode(2), 1);
  test_assert_equal(mem1_getHeapSizeOfNode(1), mem1_getSizeOfTotalHeapMem() - 2*mem1_getSizeOfHeapCtrlNode() - 100);
  test_assert_equal(mem1_getHeapSizeOfNode(2), 100);
  test_assert_equal(mem1_isHeapMemorySizeValid(), 1);
  mem1_printHeapIfFailed();

  /* free block 1 */
  mem1_free(p1);
  mem1_assertHeapEmpty();
  mem1_printHeapIfFailed();

  test_testcase_end();
}

/*******************************************************
 * mem1test_FreeMiddleTwoNext
 *
 * Steps:
 * - Allocate one block of 100
 * - Allocate one block of 200
 * - Allocate one block of 300
 * - Allocate one block of 400
 * - free block with 200
 * - free block with 300
 * - free block with 100
 * - free block with 400
 *******************************************************/
void mem1test_FreeMiddleTwoNext()
{
  void *p1, *p2, *p3, *p4;

  test_testcase_start("mem1test_FreeMiddleTwoNext");
  mem1_init();

  /* allocate block 1 */
  p1 = mem1_malloc(100);
  /* allocate block 2 */
  p2 = mem1_malloc(200);
  /* allocate block 3 */
  p3 = mem1_malloc(300);
  /* allocate block 4 */
  p4 = mem1_malloc(400);

    /* free block 2 */
  mem1_free(p2);
  test_assert_equal(mem1_getNumberOfNodes(), 5);
  test_assert_equal(mem1_getHeapUseOfNode(1), 0);
  test_assert_equal(mem1_getHeapUseOfNode(2), 1);
  test_assert_equal(mem1_getHeapUseOfNode(3), 1);
  test_assert_equal(mem1_getHeapUseOfNode(4), 0);
  test_assert_equal(mem1_getHeapUseOfNode(5), 1);
  test_assert_equal(mem1_getHeapSizeOfNode(2), 400);
  test_assert_equal(mem1_getHeapSizeOfNode(3), 300);
  test_assert_equal(mem1_getHeapSizeOfNode(4), 200);
  test_assert_equal(mem1_getHeapSizeOfNode(5), 100);
  test_assert_equal(mem1_isHeapMemorySizeValid(), 1);
  mem1_printHeapIfFailed();

  /* free block 3 */
  mem1_free(p3);
  test_assert_equal(mem1_getNumberOfNodes(), 4);
  test_assert_equal(mem1_getHeapUseOfNode(1), 0);
  test_assert_equal(mem1_getHeapUseOfNode(2), 1);
  test_assert_equal(mem1_getHeapUseOfNode(3), 0);
  test_assert_equal(mem1_getHeapUseOfNode(4), 1);
  test_assert_equal(mem1_getHeapSizeOfNode(2), 400);
  test_assert_equal(mem1_getHeapSizeOfNode(3), 300 + 200 + 16);
  test_assert_equal(mem1_getHeapSizeOfNode(4), 100);
  test_assert_equal(mem1_isHeapMemorySizeValid(), 1);
  mem1_printHeapIfFailed();

  /* free block 1 */
  mem1_free(p1);
  test_assert_equal(mem1_getNumberOfNodes(), 3);
  test_assert_equal(mem1_getHeapUseOfNode(1), 0);
  test_assert_equal(mem1_getHeapUseOfNode(2), 1);
  test_assert_equal(mem1_getHeapUseOfNode(3), 0);
  test_assert_equal(mem1_getHeapSizeOfNode(2), 400);
  test_assert_equal(mem1_getHeapSizeOfNode(3), 100 + 200 + 300 + 2*mem1_getSizeOfHeapCtrlNode());
  test_assert_equal(mem1_isHeapMemorySizeValid(), 1);
  mem1_printHeapIfFailed();

  /* free block 4 */
  mem1_free(p4);
  mem1_assertHeapEmpty();
  mem1_printHeapIfFailed();

  test_testcase_end();
}

/*******************************************************
 * mem1test_FreeMiddleAllocateAgain
 *
 * Steps:
 * - Allocate one block of heap size - 450
 * - Allocate one block of 200
 * - Allocate one block of 100
 * - free block with 200
 * - Allocate block with 150
 *******************************************************/
void mem1test_FreeMiddleAllocateAgain(void)
{
  void *p1, *p2, *p3;

  test_testcase_start("mem1test_FreeMiddleAllocateAgain");
  mem1_init();

  /* allocate block 1 */
  p1 = mem1_malloc(mem1_getSizeOfTotalHeapMem() - mem1_getSizeOfHeapCtrlNode() - 450);
  /* allocate block 2 */
  p2 = mem1_malloc(200);
  /* allocate block 3 */
  p3 = mem1_malloc(100);

  mem1_free(p2);
  test_assert_equal(mem1_getNumberOfNodes(), 4);
  test_assert_equal(mem1_getHeapUseOfNode(1), 0);
  test_assert_equal(mem1_getHeapUseOfNode(2), 1);
  test_assert_equal(mem1_getHeapUseOfNode(3), 0);
  test_assert_equal(mem1_getHeapUseOfNode(4), 1);

  test_assert_equal(mem1_getHeapSizeOfNode(1) , 450 - 3* mem1_getSizeOfHeapCtrlNode() - 200 - 100);
  test_assert_equal(mem1_getHeapSizeOfNode(3), 200);
  mem1_printHeapIfFailed();

  /* allocate block 2 with 150 */
  p2 = mem1_malloc(150);
  test_assert_not_null_ptr(p2);
  test_assert_equal(mem1_getNumberOfNodes(), 5);
  test_assert_equal(mem1_getHeapUseOfNode(1), 0);  /* heap start */
  test_assert_equal(mem1_getHeapUseOfNode(2), 1);  /* block with 100 */
  test_assert_equal(mem1_getHeapUseOfNode(3), 0);  /* remaing block from old 200 - new 150 block*/
  test_assert_equal(mem1_getHeapUseOfNode(4), 1);
  test_assert_equal(mem1_getHeapUseOfNode(5), 1);

  test_assert_equal(mem1_getHeapSizeOfNode(3), 200 - 150 - mem1_getSizeOfHeapCtrlNode());
  test_assert_equal(mem1_getHeapSizeOfNode(4), 150);

  test_testcase_end();
}

/*******************************************************
 * mem1test_ReallocLargerSize
 *
 * Steps:
 * - Allocate one block of 250
 * - Reallocate block to 300
 * - Free block
 *******************************************************/
void mem1test_ReallocLargerSize(void)
{
  void *p1, *p2;
  size_t i;

  test_testcase_start("mem1test_ReallocLargerSize");
  mem1_init();

  /* allocate block 1 */
  p1 = mem1_malloc(250);
  mem1_memset(p1, 0xFD, 250);

  /* reallocate block */
  p2 = mem1_realloc(p1, 300);
  for (i = 0; i < 250; i++)
  {
    test_assert_equal( ((uint8_t *)p2)[i], 0xFD);
  }

  /* free block */
  mem1_free(p2);
  mem1_assertHeapEmpty();
  mem1_printHeapIfFailed();

  test_testcase_end();
}

/*******************************************************
 * mem1test_ReallocSmallerSize
 *
 * Steps:
 * - Allocate one block of 450
 * - Reallocate block to 300
 * - Free block
 *******************************************************/
void mem1test_ReallocSmallerSize(void)
{
  void *p1, *p2;
  size_t i;

  test_testcase_start("mem1test_ReallocSmallerSize");
  mem1_init();

  /* allocate block 1 */
  p1 = mem1_malloc(450);
  mem1_memset(p1, 0xFA, 450);

  /* reallocate block */
  p2 = mem1_realloc(p1, 300);
  for (i = 0; i < 300; i++)
  {
    test_assert_equal( ((uint8_t *)p2)[i], 0xFA);
  }

  /* free block */
  mem1_free(p2);
  mem1_assertHeapEmpty();
  mem1_printHeapIfFailed();

  test_testcase_end();
}

#endif