/******************************************************************************************************************************************
 *
 * Filename    : mem2.c
 * Project     : MemAlloc
 * Description : This is an example implementation of the C memory functions malloc(), free(), calloc() and realloc().
 *               It should serve as an tutorial for learning - it's not an efficient, optimized implementation.
 *
 *               See the corresponding article on my homepage for explanation of the algorithms.
 *
 *               This file contains the implementation variant 2 of the C memory functions.
 *
 * Sunshine, October 2k19
 * www.sunshine2k.de | www.bastian-molkenthin.de
 *
 * This file is licensed under MIT licence. For further details, see http://www.sunshine2k.de/license.html.
 *
 *****************************************************************************************************************************************/

/******************************************************************************************************************************************
 * INCLUDES
 *****************************************************************************************************************************************/

#include "mem2.h"
#if defined ( MEM2_TEST_MODE )
# include <stdio.h>
# include "testutil.h"
#endif

/******************************************************************************************************************************************
 * DEFINES
 *****************************************************************************************************************************************/
#define HEAP_TOTAL_SIZE       4*1024
#define HEAP_NODE_SIZE        sizeof(MyHeapNodeType2)

/******************************************************************************************************************************************
 * TYPES
 *****************************************************************************************************************************************/
typedef struct MyHeapNodeTypeTag2
{
  uint32_t size;                              /* size of heap, without size of control data */
  struct MyHeapNodeTypeTag2* nextNode;         /* pointer to next heap entry */
} MyHeapNodeType2;

/******************************************************************************************************************************************
 * VARIABLES
 *****************************************************************************************************************************************/
static unsigned char MyHeapArea[HEAP_TOTAL_SIZE];
static MyHeapNodeType2* heapStart = (MyHeapNodeType2 *)MyHeapArea;

/******************************************************************************************************************************************
 * LOCAL FUNCTION DECLARATIONS
 *****************************************************************************************************************************************/

/******************************************************************************************************************************************
 * FUNCTIONS
 *****************************************************************************************************************************************/

/**************************************************************************************************
 * mem2_init
 **************************************************************************************************/
void mem2_init()
{
  heapStart = (MyHeapNodeType2 *)MyHeapArea;
  heapStart->size = HEAP_TOTAL_SIZE - HEAP_NODE_SIZE;
  heapStart->nextNode = NULL_PTR;
}

/**************************************************************************************************
 * mem2_malloc
 **************************************************************************************************/
void * mem2_malloc(size_t size)
{
  MyHeapNodeType2* currentHeapBlock;
  MyHeapNodeType2* bestHeapBlock;
  uint32_t bestHeapBlockSize;

  /* init current block to start of heap */
  currentHeapBlock = heapStart;
  /* init best heap block */
  bestHeapBlock = (MyHeapNodeType2*)NULL_PTR;
  bestHeapBlockSize = HEAP_TOTAL_SIZE;

  while (currentHeapBlock)
  {
    /* check if current block matches and fittest best (so far) */
    if ( (currentHeapBlock->size >= (size + HEAP_NODE_SIZE)) && 
         (currentHeapBlock->size <= bestHeapBlockSize) )
    {
      bestHeapBlock = currentHeapBlock;
      bestHeapBlockSize = currentHeapBlock->size;
    }

    currentHeapBlock = currentHeapBlock->nextNode;
  }

  if (bestHeapBlock != NULL_PTR)
  {
    MyHeapNodeType2* heapNodeAllocate;

    /* found a matching block, split it up and return the top of the memory area to the user */
    /* the best matching block is decreased by the needed memory area */
    bestHeapBlock->size = bestHeapBlock->size - size - HEAP_NODE_SIZE;
    /* new heap node is after the current heap + the size of its control data + allocated memory size */
    heapNodeAllocate = (MyHeapNodeType2*)(((unsigned char*)bestHeapBlock) + HEAP_NODE_SIZE + bestHeapBlock->size);
    heapNodeAllocate->size = size;
    /* return pointer to memory of new heap node after control data */
    return (void*)((unsigned char*)heapNodeAllocate + HEAP_NODE_SIZE);
  }

  return NULL_PTR;
}

/**************************************************************************************************
 * mem2_free
 **************************************************************************************************/
void mem2_free(void* p)
{
  MyHeapNodeType2* prevBlock;
  MyHeapNodeType2* nextBlock;

  if (p == NULL_PTR)
  {
    return;
  }

  /* get actual heap node */
  MyHeapNodeType2* blockToFree = (MyHeapNodeType2*)((unsigned char*)p - HEAP_NODE_SIZE);

  if (blockToFree == NULL_PTR)
  {
    return;
  }

  /* start at beginning of heap, find neighbor block with lower address*/
  prevBlock = NULL_PTR;
  nextBlock = heapStart;
  while ((nextBlock != NULL_PTR) && (nextBlock < blockToFree))
  {
    prevBlock = nextBlock;
    nextBlock = nextBlock->nextNode;
  }

  /* add current block that shall be freed back to heap list */
  blockToFree->nextNode = nextBlock;
  if (prevBlock != NULL_PTR)
  {
    prevBlock->nextNode = blockToFree;
  }

  /* check if block can be merged with right block */
  if ( (nextBlock != NULL_PTR) && 
       ( ((unsigned char*)blockToFree + blockToFree->size + HEAP_NODE_SIZE) == (unsigned char*)nextBlock) )
  {
    blockToFree->size += nextBlock->size + HEAP_NODE_SIZE;
    blockToFree->nextNode = nextBlock->nextNode;
  }

  /* check if block can be merged with left block */
  if ( (prevBlock != NULL_PTR) && 
     ( ((unsigned char*)prevBlock + prevBlock->size + HEAP_NODE_SIZE) == (unsigned char*)blockToFree) )
  {
    prevBlock->size += blockToFree->size + HEAP_NODE_SIZE;
    prevBlock->nextNode = blockToFree->nextNode;
  }
}

/**************************************************************************************************
 * mem2_calloc
 **************************************************************************************************/
void* mem2_calloc(size_t num, size_t size)
{
  void* p;
  size_t sizeInBytes;

  /* caluclate actual memory size in bytes */
  sizeInBytes = num * size;
  /* allocate memory block */
  p = mem2_malloc(sizeInBytes);
  /* initialize memory block to zero */
  if (p != NULL_PTR)
  {
    mem2_memset(p, 0, sizeInBytes);
  }
  return p;
}

/**************************************************************************************************
 * mem2_realloc
 **************************************************************************************************/
void* mem2_realloc(void* ptr, size_t size)
{
  void* p;

  /* allocate new block of requested size */
  p = mem2_malloc(size);
  if (p == NULL_PTR)
  {
    return p; /* return NULL_PTR in case of failure, not original ptr! */
  }

  /* copy buffer of original block */
  if (ptr != NULL_PTR)
  {
    /* get original size of memory block */
    size_t sizeToCopy;
    size_t originalSize = ((MyHeapNodeType2* )ptr)->size;

    /* copy data of original memory block to new one - only valid data is copied */
    if (originalSize < size)
    {
      sizeToCopy = originalSize;
    }
    else
    {
      sizeToCopy = size;
    }
    mem2_memcpy(p, ptr, sizeToCopy);

    /* free original block */
    mem2_free(ptr);
  }

  return p;
}


/**************************************************************************************************
 * mem2_memset
 **************************************************************************************************/
void* mem2_memset(void * ptr, int value, size_t num)
{
  uint8_t* p = (uint8_t*) ptr;
  while (num > 0)
  {
    *p = value;
    p++;
    num--;
  }
  return ptr;
}

/**************************************************************************************************
 * mem2_memcpy
 **************************************************************************************************/
void* mem2_memcpy(void * dest, const void * source, size_t num)
{
  size_t i;
  uint8_t* pDest = (uint8_t*)dest;
  uint8_t* pSource = (uint8_t*)source;

  for (i = 0; i < num; i++)
  {
    pDest[i] = pSource[i];
  }

  return dest;
}




#if defined (MEM2_TEST_MODE)
/**************************************************************************************************
 * mem2_printHeapheader
 **************************************************************************************************/
void mem2_printHeapheader()
{
  printf("====================================================\n");
  printf("Start of Heap: 0x%08X\n", (unsigned int)heapStart);
  printf("Size of Heap : 0x%08X (%d)\n", HEAP_TOTAL_SIZE - HEAP_NODE_SIZE, HEAP_TOTAL_SIZE - HEAP_NODE_SIZE);
  printf("====================================================\n");
}

/**************************************************************************************************
 * mem2_printHeap
 **************************************************************************************************/
void mem2_printHeap()
{
  MyHeapNodeType2* currentHeapBlock;
  printf("\n--------------------------------\n");
  printf("Start\t\tSize\t\t\tNext\n");

  currentHeapBlock = heapStart;
  while (currentHeapBlock)
  {
    printf("0x%08X\t0x%08X (%d)\t0x%08X \n", 
      (unsigned int)currentHeapBlock, currentHeapBlock->size, currentHeapBlock->size,
      (unsigned int)currentHeapBlock->nextNode);

    currentHeapBlock = currentHeapBlock->nextNode;
  }
}

/**************************************************************************************************
 * mem2_printHeapIfFailed
 **************************************************************************************************/
void mem2_printHeapIfFailed(void)
{
  if (test_get_testcase_failed() != 0)
  {
    mem2_printHeap();
  }
}

/**************************************************************************************************
 * mem2_getSizeOfHeapCtrlNode
 **************************************************************************************************/
uint32_t  mem2_getSizeOfHeapCtrlNode()
{
  return HEAP_NODE_SIZE;
}

/**************************************************************************************************
 * mem2_getSizeOfTotalHeapMem
 **************************************************************************************************/
uint32_t  mem2_getSizeOfTotalHeapMem()
{
  return HEAP_TOTAL_SIZE;
}

/**************************************************************************************************
 * mem2_getHeapStart
 **************************************************************************************************/
void* mem2_getHeapStart()
{
  return (void*)heapStart;
}

/**************************************************************************************************
 * mem2_getNumberOfNodes
 **************************************************************************************************/
uint8_t   mem2_getNumberOfNodes(void)
{
  uint8_t numOfNodes = 1;
  MyHeapNodeType2* node = heapStart;

  while (node->nextNode != NULL_PTR)
  {
    numOfNodes++;
    node = node->nextNode;
  }

  return numOfNodes;
}

/**************************************************************************************************
 * mem2_getHeapSizeOfNode
 **************************************************************************************************/
uint32_t   mem2_getHeapSizeOfNode(uint8_t numOfNode)
{
  uint8_t currentNodeNum = 1;
  MyHeapNodeType2* node = heapStart;

  while (currentNodeNum < numOfNode && node->nextNode != NULL_PTR)
  {
    currentNodeNum++;
    node = node->nextNode;
  }

  if (currentNodeNum != numOfNode)
  {
    return 0;
  }

  return node->size;
}

/**************************************************************************************************
 * mem2_isHeapMemorySizeValid
 * Only works for empty heap!
 **************************************************************************************************/
uint8_t mem2_isHeapMemorySizeValid(void)
{
  uint8_t numOfNodes = 0;
  MyHeapNodeType2* node = heapStart;
  uint32_t size = 0;

  do
  {
    numOfNodes++;
    size += node->size;
    node = node->nextNode;
  }
  while (node != NULL_PTR);

  size = size + (numOfNodes * HEAP_NODE_SIZE);

  if (size == HEAP_TOTAL_SIZE)
  {
    return 1;
  }
  else
  {
    return 0;
  }
}

/**************************************************************************************************
 * mem2_getEmptyHeapSpace
 * Returns the total remaining heap space
 **************************************************************************************************/
uint32_t  mem2_getEmptyHeapSpace()
{
  MyHeapNodeType2* node = heapStart;
  uint32_t size = 0;

  do
  {
    size += node->size;
    node = node->nextNode;
  }
  while (node != NULL_PTR);

  return size;
}


#endif
