/******************************************************************************************************************************************
 *
 * Filename    : mem2test.c
 * Project     : MemAlloc
 * Description : This is an example implementation of the C memory functions malloc(), free(), calloc() and realloc().
 *               It should serve as an tutorial for learning - it's not an efficient, optimized implementation.
 *
 *               See the corresponding article on my homepage for explanation of the algorithms.
 *
 *               This file contains the test cases for implementation variant 2 C memory functions.
 *
 * Sunshine, October 2k19
 * www.sunshine2k.de | www.bastian-molkenthin.de
 *
 * This file is licensed under MIT licence. For further details, see http://www.sunshine2k.de/license.html.
 *
 *****************************************************************************************************************************************/

/******************************************************************************************************************************************
 * INCLUDES
 *****************************************************************************************************************************************/
#include "mem2test.h"
#include "testutil.h"

/******************************************************************************************************************************************
 * DEFINES
 *****************************************************************************************************************************************/

/******************************************************************************************************************************************
 * TYPES
 *****************************************************************************************************************************************/

/******************************************************************************************************************************************
 * LOCAL FUNCTION DECLARATIONS
 *****************************************************************************************************************************************/
#if defined ( MEM2_TEST_MODE )
static void mem2_assertHeapEmpty();
#endif

/******************************************************************************************************************************************
 * LOCAL FUNCTIONS
 *****************************************************************************************************************************************/
#if defined ( MEM2_TEST_MODE )
static void mem2_assertHeapEmpty()
{
  test_assert_equal(mem2_getNumberOfNodes(), 1);
  test_assert_equal(mem2_getHeapSizeOfNode(1), mem2_getSizeOfTotalHeapMem() - mem2_getSizeOfHeapCtrlNode());
  test_assert_equal(mem2_isHeapMemorySizeValid(), 1);
}
#endif

/******************************************************************************************************************************************
 * FUNCTIONS
 *****************************************************************************************************************************************/

#if defined ( MEM2_TEST_MODE )
/*******************************************************
 * mem2test_FullHeapAlloc
 *
 * Steps:
 * - Allocate full heap and assert success
 * - Allocate another block and assert that it fails
 * - Free allocated block and assert that full heap is free
 *******************************************************/
void mem2test_FullHeapAlloc(void)
{
  void* p, *pFail;

  test_testcase_start("mem2test_FullHeapAlloc");
  mem2_init();
  test_assert_equal(mem2_isHeapMemorySizeValid(), 1);
  mem2_assertHeapEmpty();
  test_assert_equal(mem2_getEmptyHeapSpace(), mem2_getSizeOfTotalHeapMem() - mem2_getSizeOfHeapCtrlNode());

  /* allocate complete heap */
  p = mem2_malloc(mem2_getSizeOfTotalHeapMem() - 2 * mem2_getSizeOfHeapCtrlNode());
  test_assert_not_null_ptr(p);
  mem2_printHeapIfFailed();
  test_assert_equal(mem2_getEmptyHeapSpace(), 0);

  /* get another heap memory block and check that it's fail */
  pFail = mem2_malloc(10);
  test_assert_null_ptr(pFail);

  /* free complete allocated memory block */
  mem2_free(p);
  mem2_printHeapIfFailed();
  mem2_assertHeapEmpty();
  test_assert_equal(mem2_isHeapMemorySizeValid(), 1);

  test_testcase_end();
}

/*******************************************************
 * mem2test_FullHeapAllocInSecondRequest
 *
 * Steps:
 * - Allocate one block of 100
 * - Allocate remaining heap
 * - free block with remaining heap
 * - free block with 100
 *******************************************************/
void mem2test_FullHeapAllocInSecondRequest(void)
{
  void* p1, *p2;

  test_testcase_start("mem2test_FullHeapAllocInSecondRequest");
  mem2_init();

  /* allocate block 1 */
  p1 = mem2_malloc(100);
  test_assert_not_null_ptr(p1);
  test_assert_equal(mem2_getEmptyHeapSpace(), mem2_getSizeOfTotalHeapMem() - 2 * mem2_getSizeOfHeapCtrlNode() - 100);

  /* allocate remaining heap */
  p2 = mem2_malloc(mem2_getSizeOfTotalHeapMem() - 3*mem2_getSizeOfHeapCtrlNode() - 100);
  test_assert_not_null_ptr(p2);
  test_assert_equal(mem2_getEmptyHeapSpace(), 0);
  mem2_printHeapIfFailed();

  /* free remaining block */
  mem2_free(p2);
  test_assert_equal(mem2_getEmptyHeapSpace(), mem2_getSizeOfTotalHeapMem() - 2 * mem2_getSizeOfHeapCtrlNode() - 100);
  mem2_printHeapIfFailed();

  /* free block 1 */
  mem2_free(p1);
  mem2_assertHeapEmpty();

  test_testcase_end();
}

/*******************************************************
 * mem2test_FullHeapAllocInSecondRequest_FreeReversedOrder
 *
 * Steps:
 * - Allocate one block of 100
 * - Allocate remaining heap
 * - free block with 100
 * - free block with remaining heap
 *******************************************************/
void mem2test_FullHeapAllocInSecondRequest_FreeReversedOrder(void)
{
  void* p1, *p2;

  test_testcase_start("mem2test_FullHeapAllocInSecondRequest_FreeReversedOrder");
  mem2_init();

  /* allocate block 1 */
  p1 = mem2_malloc(100);

  /* allocate remaining heap */
  p2 = mem2_malloc(mem2_getSizeOfTotalHeapMem() - 3*mem2_getSizeOfHeapCtrlNode() - 100);
  test_assert_not_null_ptr(p2);
  test_assert_equal(mem2_getEmptyHeapSpace(), 0);
  mem2_printHeapIfFailed();

  /* free block 1 */
  mem2_free(p1);
  test_assert_equal(mem2_getEmptyHeapSpace(), 100);
  mem2_printHeapIfFailed();

  /* free block 2 */
  mem2_free(p2);
  mem2_assertHeapEmpty();
  test_assert_equal(mem2_getEmptyHeapSpace(), mem2_getSizeOfTotalHeapMem() - mem2_getSizeOfHeapCtrlNode());
  mem2_printHeapIfFailed();

  test_testcase_end();
}

/*******************************************************
 * mem2test_AllocRequestSizeTooLarge
 *
 * Steps:
 * - Allocate block with too large size
 *******************************************************/
void mem2test_AllocRequestSizeTooLarge(void)
{
  void* p1, *p2;

  test_testcase_start("mem2test_AllocRequestSizeTooLarge");
  mem2_init();

  /* try to allocate block with too large requested size */
  p1 = mem2_malloc(mem2_getSizeOfTotalHeapMem() - 2 * mem2_getSizeOfHeapCtrlNode() + 1);
  test_assert_null_ptr(p1);
  mem2_assertHeapEmpty();

  /* allocate valid block */
  p1 = mem2_malloc(99);
  test_assert_equal(mem2_getEmptyHeapSpace(), mem2_getSizeOfTotalHeapMem() - 2 * mem2_getSizeOfHeapCtrlNode() - 99);
  mem2_printHeapIfFailed();

  /* try to allocate block with too large requested size */
  p2 = mem2_malloc(mem2_getSizeOfTotalHeapMem() - 3 * mem2_getSizeOfHeapCtrlNode() -99 + 1);
  test_assert_null_ptr(p2);
  test_assert_equal(mem2_getEmptyHeapSpace(), mem2_getSizeOfTotalHeapMem() - 2 * mem2_getSizeOfHeapCtrlNode() - 99);
  mem2_printHeapIfFailed();

  /* free valid block */
  mem2_free(p1);
  mem2_assertHeapEmpty();
  mem2_printHeapIfFailed();

  test_testcase_end();
}

/*******************************************************
 * mem2test_FreeBasic
 *
 * Steps:
 * - Allocate one block of 1000
 * - free block with 1000
 *******************************************************/
void mem2test_FreeBasic(void)
{
  void* p1;

  test_testcase_start("mem2test_FreeBasic");
  mem2_init();

  /* allocate block 1 */
  p1 = mem2_malloc(1000);
  test_assert_not_null_ptr(p1);
  test_assert_equal(mem2_getEmptyHeapSpace(), mem2_getSizeOfTotalHeapMem() - 2 * mem2_getSizeOfHeapCtrlNode() - 1000);
  mem2_printHeapIfFailed();

  /* free block 1 */
  mem2_free(p1);
  mem2_assertHeapEmpty();
  mem2_printHeapIfFailed();

  test_testcase_end();
}

/*******************************************************
 * mem2test_FreeLatestToFirst
 *
 * Steps:
 * - Allocate one block of 100
 * - Allocate one block of 200
 * - Allocate one block of 300
 * - free block with 300
 * - free block with 200
 * - free block with 100
 *******************************************************/
void mem2test_FreeLatestToFirst()
{
  void *p1, *p2, *p3;

  test_testcase_start("mem2test_FreeLatestToFirst");
  mem2_init();

  /* allocate block 1 */
  p1 = mem2_malloc(100);
  test_assert_not_null_ptr(p1);
  test_assert_equal(mem2_getEmptyHeapSpace(), mem2_getSizeOfTotalHeapMem() - 2 * mem2_getSizeOfHeapCtrlNode() - 100);
  mem2_printHeapIfFailed();

  /* allocate block 2 */
  p2 = mem2_malloc(200);
  test_assert_equal(mem2_getEmptyHeapSpace(), mem2_getSizeOfTotalHeapMem() - 3 * mem2_getSizeOfHeapCtrlNode() - 100 - 200);
  mem2_printHeapIfFailed();

  /* allocate block 3 */
  p3 = mem2_malloc(300);
  test_assert_not_null_ptr(p3);
  test_assert_equal(mem2_getEmptyHeapSpace(), mem2_getSizeOfTotalHeapMem() - 4 * mem2_getSizeOfHeapCtrlNode() - 100 - 200 - 300);
  mem2_printHeapIfFailed();

  /* free block 3 */
  mem2_free(p3);
  test_assert_equal(mem2_getEmptyHeapSpace(), mem2_getSizeOfTotalHeapMem() - 3 * mem2_getSizeOfHeapCtrlNode() - 100 - 200);
  mem2_printHeapIfFailed();

  /* free block 2 */
  mem2_free(p2);
  test_assert_equal(mem2_getEmptyHeapSpace(), mem2_getSizeOfTotalHeapMem() - 2 * mem2_getSizeOfHeapCtrlNode() - 100);
  mem2_printHeapIfFailed();

  /* free block 1*/
  mem2_free(p1);
  mem2_assertHeapEmpty();
  mem2_printHeapIfFailed();

  test_testcase_end();
}


/*******************************************************
 * mem2test_FreeFirstToLatest
 *
 * Steps:
 * - Allocate one block of 100
 * - Allocate one block of 200
 * - Allocate one block of 300
 * - free block with 100
 * - free block with 200
 * - free block with 300
 *******************************************************/
void mem2test_FreeFirstToLatest()
{
  void *p1, *p2, *p3;

  test_testcase_start("mem2test_FreeFirstToLatest");
  mem2_init();

  /* allocate block 1 */
  p1 = mem2_malloc(100);
  test_assert_not_null_ptr(p1);
  test_assert_equal(mem2_getEmptyHeapSpace(), mem2_getSizeOfTotalHeapMem() - 2 * mem2_getSizeOfHeapCtrlNode() - 100);
  mem2_printHeapIfFailed();

  /* allocate block 2 */
  p2 = mem2_malloc(200);
  test_assert_not_null_ptr(p2);
  test_assert_equal(mem2_getEmptyHeapSpace(), mem2_getSizeOfTotalHeapMem() - 3 * mem2_getSizeOfHeapCtrlNode() - 100 - 200);
  mem2_printHeapIfFailed();

  /* allocate block 3 */
  p3 = mem2_malloc(300);
  test_assert_not_null_ptr(p3);
  test_assert_equal(mem2_getEmptyHeapSpace(), mem2_getSizeOfTotalHeapMem() - 4 * mem2_getSizeOfHeapCtrlNode() - 100 - 200 - 300);
  mem2_printHeapIfFailed();

  /* free block 1 */
  mem2_free(p1);
  test_assert_equal(mem2_getEmptyHeapSpace(), mem2_getSizeOfTotalHeapMem() - 4 * mem2_getSizeOfHeapCtrlNode() - 200 - 300);
  mem2_printHeapIfFailed();

  /* free block 2 */
  mem2_free(p2);
  test_assert_equal(mem2_getEmptyHeapSpace(), mem2_getSizeOfTotalHeapMem() - 3 * mem2_getSizeOfHeapCtrlNode() - 300);
  mem2_printHeapIfFailed();

  /* free block 3 */
  mem2_free(p3);
  mem2_assertHeapEmpty();
  mem2_printHeapIfFailed();

  test_testcase_end();
}

/*******************************************************
 * mem2test_FreeMiddlePrevious
 *
 * Steps:
 * - Allocate one block of 100
 * - Allocate one block of 200
 * - Allocate one block of 300
 * - free block with 200
 * - free block with 300
 * - free block with 100
 *******************************************************/
void mem2test_FreeMiddlePrevious()
{
  void *p1, *p2, *p3;

  test_testcase_start("mem2test_FreeMiddlePrevious");
  mem2_init();

  /* allocate block 1 */
  p1 = mem2_malloc(100);
  /* allocate block 2 */
  p2 = mem2_malloc(200);
  /* allocate block 3 */
  p3 = mem2_malloc(300);

  /* free block 2 */
  mem2_free(p2);
  test_assert_equal(mem2_getEmptyHeapSpace(), mem2_getSizeOfTotalHeapMem() - 4 * mem2_getSizeOfHeapCtrlNode() - 300 - 100);
  mem2_printHeapIfFailed();

  /* free block 3 */
  mem2_free(p3);
  test_assert_equal(mem2_getEmptyHeapSpace(), mem2_getSizeOfTotalHeapMem() - 2 * mem2_getSizeOfHeapCtrlNode() - 100);
  mem2_printHeapIfFailed();

  /* free block 1 */
  mem2_free(p1);
  mem2_assertHeapEmpty();
  mem2_printHeapIfFailed();

  test_testcase_end();
}

/*******************************************************
 * mem2test_FreeMiddleNext
 *
 * Steps:
 * - Allocate one block of 100
 * - Allocate one block of 200
 * - Allocate one block of 300
 * - free block with 200
 * - free block with 100
 * - free block with 300
 *******************************************************/
void mem2test_FreeMiddleNext()
{
  void *p1, *p2, *p3;

  test_testcase_start("mem2test_FreeMiddleNext");
  mem2_init();

  /* allocate block 1 */
  p1 = mem2_malloc(100);
  /* allocate block 2 */
  p2 = mem2_malloc(200);
  /* allocate block 3 */
  p3 = mem2_malloc(300);

  /* free block 2 */
  mem2_free(p2);
  test_assert_equal(mem2_getEmptyHeapSpace(), mem2_getSizeOfTotalHeapMem() - 4 * mem2_getSizeOfHeapCtrlNode() - 300 - 100);
  mem2_printHeapIfFailed();

  /* free block 1 */
  mem2_free(p1);
  test_assert_equal(mem2_getEmptyHeapSpace(), mem2_getSizeOfTotalHeapMem() - 3 * mem2_getSizeOfHeapCtrlNode() - 300);
  mem2_printHeapIfFailed();

  /* free block 3 */
  mem2_free(p3);
  mem2_assertHeapEmpty();
  mem2_printHeapIfFailed();

  test_testcase_end();
}

/*******************************************************
 * mem2test_FreeMiddleTwoPrevious
 *
 * Steps:
 * - Allocate one block of 100
 * - Allocate one block of 200
 * - Allocate one block of 300
 * - Allocate one block of 400
 * - free block with 200
 * - free block with 300
 * - free block with 400
 * - free block with 100
 *******************************************************/
void mem2test_FreeMiddleTwoPrevious()
{
  void *p1, *p2, *p3, *p4;

  test_testcase_start("mem2test_FreeMiddleTwoPrevious");
  mem2_init();

  /* allocate block 1 */
  p1 = mem2_malloc(100);
  /* allocate block 2 */
  p2 = mem2_malloc(200);
  /* allocate block 3 */
  p3 = mem2_malloc(300);
  /* allocate block 4 */
  p4 = mem2_malloc(400);
  test_assert_not_null_ptr(p4);
  test_assert_equal(mem2_getEmptyHeapSpace(), mem2_getSizeOfTotalHeapMem() - 5 * mem2_getSizeOfHeapCtrlNode() - 100 - 200 - 300 - 400);
  mem2_printHeapIfFailed();

  /* free block 2 */
  mem2_free(p2);
  test_assert_equal(mem2_getEmptyHeapSpace(), mem2_getSizeOfTotalHeapMem() - 5 * mem2_getSizeOfHeapCtrlNode() - 100 - 300 - 400);
  mem2_printHeapIfFailed();

  /* free block 3 */
  mem2_free(p3);
  test_assert_equal(mem2_getEmptyHeapSpace(), mem2_getSizeOfTotalHeapMem() - 4 * mem2_getSizeOfHeapCtrlNode() - 100 - 400);
  mem2_printHeapIfFailed();

  /* free block 4 */
  mem2_free(p4);
  test_assert_equal(mem2_getEmptyHeapSpace(), mem2_getSizeOfTotalHeapMem() - 2 * mem2_getSizeOfHeapCtrlNode() - 100);
  mem2_printHeapIfFailed();

  /* free block 1 */
  mem2_free(p1);
  mem2_assertHeapEmpty();
  mem2_printHeapIfFailed();

  test_testcase_end();
}

/*******************************************************
 * mem2test_FreeMiddleTwoNext
 *
 * Steps:
 * - Allocate one block of 100
 * - Allocate one block of 200
 * - Allocate one block of 300
 * - Allocate one block of 400
 * - free block with 200
 * - free block with 300
 * - free block with 100
 * - free block with 400
 *******************************************************/
void mem2test_FreeMiddleTwoNext()
{
  void *p1, *p2, *p3, *p4;

  test_testcase_start("mem2test_FreeMiddleTwoNext");
  mem2_init();

  /* allocate block 1 */
  p1 = mem2_malloc(100);
  /* allocate block 2 */
  p2 = mem2_malloc(200);
  /* allocate block 3 */
  p3 = mem2_malloc(300);
  /* allocate block 4 */
  p4 = mem2_malloc(400);

  /* free block 2 */
  mem2_free(p2);
  test_assert_equal(mem2_getEmptyHeapSpace(), mem2_getSizeOfTotalHeapMem() - 5 * mem2_getSizeOfHeapCtrlNode() - 100 - 300 - 400);
  mem2_printHeapIfFailed();

  /* free block 3 */
  mem2_free(p3);
  test_assert_equal(mem2_getEmptyHeapSpace(), mem2_getSizeOfTotalHeapMem() - 4 * mem2_getSizeOfHeapCtrlNode() - 100 - 400);
  mem2_printHeapIfFailed();

  /* free block 1 */
  mem2_free(p1);
  test_assert_equal(mem2_getEmptyHeapSpace(), mem2_getSizeOfTotalHeapMem() - 3 * mem2_getSizeOfHeapCtrlNode() - 400);
  mem2_printHeapIfFailed();

  /* free block 4 */
  mem2_free(p4);
  mem2_assertHeapEmpty();
  mem2_printHeapIfFailed();

  test_testcase_end();
}

/*******************************************************
 * mem2test_ReallocLargerSize
 *
 * Steps:
 * - Allocate one block of 250
 * - Reallocate block to 300
 * - Free block
 *******************************************************/
void mem2test_ReallocLargerSize(void)
{
  void *p1, *p2;
  size_t i;

  test_testcase_start("mem2test_ReallocLargerSize");
  mem2_init();

  /* allocate block 1 */
  p1 = mem2_malloc(250);
  mem2_memset(p1, 0xFD, 250);

  /* reallocate block */
  p2 = mem2_realloc(p1, 300);
  for (i = 0; i < 250; i++)
  {
    test_assert_equal( ((uint8_t *)p2)[i], 0xFD);
  }

  /* free block */
  mem2_free(p2);
  mem2_assertHeapEmpty();
  mem2_printHeapIfFailed();

  test_testcase_end();
}

/*******************************************************
 * mem2test_ReallocSmallerSize
 *
 * Steps:
 * - Allocate one block of 250
 * - Reallocate block to 300
 * - Free block
 *******************************************************/
void mem2test_ReallocSmallerSize(void)
{
  void *p1, *p2;
  size_t i;

  test_testcase_start("mem2test_ReallocSmallerSize");
  mem2_init();

  /* allocate block 1 */
  p1 = mem2_malloc(450);
  mem2_memset(p1, 0xFA, 450);

  /* reallocate block */
  p2 = mem2_realloc(p1, 300);
  for (i = 0; i < 300; i++)
  {
    test_assert_equal( ((uint8_t *)p2)[i], 0xFA);
  }

  /* free block */
  mem2_free(p2);
  mem2_assertHeapEmpty();
  mem2_printHeapIfFailed();

  test_testcase_end();
}

#endif
